--                              -*- Mode: Ada -*-
-- Filename        : $Source: /home/erdmann/ASCL/Library/RCS/ascl-sv-server.adb,v $
-- Description     : server component
-- Author          : Michael Erdmann
-- Created On      : Fri Jul 24 21:48:50 1998
-- Last Modified By: $Author: erdmann $
-- Last Modified On: $Date: 1999/09/28 13:08:43 $
-- Update Count    : 0
-- Status          : $State: Exp $
--
-- Functional Description
-- ======================
-- This is the server factory. For each incoming socket a new
-- agent process is created. The agent process reads in text lines
-- processes them and returns a text string to the client.
--
--
-- History
-- =======
-- $Log: ascl-sv-server.adb,v $
-- Revision 1.5  1999/09/28 13:08:43  erdmann
-- Tests completed a patch in the florist package is needed
--
-- Revision 1.4  1999/09/28 10:01:19  erdmann
-- Switched to posix bindings
--
-- Revision 1.3  1999/09/28 07:40:05  erdmann
-- Debugging Support added
--
-- Revision 1.2  1999/09/26 18:18:19  erdmann
-- Changed interface to process. Several output packages possible
--
-- Revision 1.1  1999/09/25 19:37:12  erdmann
-- No comments
--
--
--
with Ada.Exceptions;               use Ada.Exceptions;
with Ada.Synchronous_Task_Control; use Ada.Synchronous_Task_Control;
with Unchecked_Deallocation;
with Unchecked_Conversion;

with POSIX;                        use POSIX;
with POSIX.IO;                     use POSIX.IO;
with POSIX.Files;                  use POSIX.Files;
with POSIX.Sockets;                use POSIX.Sockets;
with POSIX.Sockets.Internet;       use POSIX.Sockets.Internet;
with GNAT.IO;                      use GNAT.IO;
with Ada.Streams;                  use Ada.Streams;

with ASCL.SV.Agent;                use ASCL.SV.Agent;
with ASCL.Debugging_Support;       use ASCL.Debugging_Support;
with ASCL;

package body ASCL.SV.Server is

   Module : constant String := "ascl-sv.server.adb";
   Max_Buffer_Size : constant Positive := 512;

   task type Factory_Task( This : Object_Data_Access );
   type Factory_Task_Access is access Factory_Task;

   -- this ist the per server control block
   type Object_Data is record
         Self                : Server.Reference := null;
         Listening_Socket    : File_Descriptor;
         Port_Number         : Internet_Port;

         Factory             : Factory_Task_Access;
         Debug               : Debugging_Support.Handle := null;
      end record;

   ---======================================================================---
   ---==== NON VISIBLE METHODS =============================================---
   ---======================================================================---

   function Initialize( Port : in Positive ) return Object_Data_Access is
      Result : Object_Data_Access := new Object_Data;
   begin
      Result.Port_Number := Internet_Port( Port );
      return Result;
   end Initialize;

   --|
   --| This is the main factory loop for the server instance. This
   --| method listens on a given port number and creates agents
   --| for a new connection.
   --|
   --| The factory task maintains a agent control list (ACL) which
   --| contains all agents which have been created and still
   --| allive.
   --|
   task body Factory_Task is
      Incomming_Socket  : File_Descriptor;
      Running           : Boolean := True;        -- factory is running
      Agent_Number      : Integer;                -- number of the currently
                                                  -- handled agent.
      -- Agent control list management
      task type Agent_Task( Nbr : Integer );
      type Agent_Task_Access is access Agent_Task;

      type ACB is record
         Agent_Ref   : Agent.Reference          := null;
         IO_Socket   : File_Descriptor;
         Process     : Agent_Task_Access        := null;
         Debug       : Debugging_Support.Handle := null;
      end record;

      Max_Agents     : Positive := 50;
      To_Many_Agents : exception;

      type Agent_Control_List is array(1..Max_Agents) of ACB;

      ACL            : Agent_Control_List ;

      --
      -- Execute the agent task. The agent task reads data from
      -- the socket and returns a string to the sender.
      --
      task body Agent_Task is
         This   : ACB renames ACL(Nbr);
         Self   : Agent.Object renames ACL(Nbr).Agent_Ref.all;
         Buffer : String (1..Max_Buffer_Size);
         Last   : POSIX.IO_Count;

         procedure Send( Message : in String ) is
         begin
            Send (This.IO_Socket,
                  Message'Address,
                  Message'Length,
                  Last);
         end Send;
      begin
         Enter( This.debug, Module & ".Agent_Task" );
         Initialize( Self );

         loop
            Receive (This.IO_Socket,
                     Buffer (Buffer'First)'Address,
                     POSIX.IO_Count (Buffer'Length),
                     Last);

            Process( Self, Buffer(1..Positive(Last)) );

            while Data_Available( Self ) loop
               Send( Get( Self ) );
            end loop;
         end loop;

      exception
         when The_Error : others =>
            Leave( This.Debug, "Some exception");
            Close( This.IO_Socket );
            Finalize( self );
            ACL(Nbr).Process := null;

      end Agent_Task;

   begin
      Enter( This.debug, Module & ".Factory_Task" );

      Running := True;
      while Running loop
         begin
            Trace( This.Debug, "Waiting for connections");
            Incomming_Socket := Accept_Connection (This.Listening_Socket);
            Trace( This.Debug, "Connection Requested");
            Agent_Number := 0;
            for I in 1..Max_Agents loop
               if ACL(I).Process  = null then
                  Agent_Number := I;
                  exit;
               end if;
            end loop;

            if Agent_Number = 0 then
               raise To_Many_Agents;
            end if;

            ACL(Agent_Number).Agent_Ref := Create( This.Self.all );
            ACL(Agent_Number).IO_Socket := Incomming_Socket;
            ACL(Agent_Number).Debug     := This.Debug;

            ACL(Agent_Number).Process   := new Agent_Task(Agent_Number);

         exception
            when To_Many_Agents =>
               null;
            when The_Error : others =>
               Error( This.Debug,
                       "*** Exception " & Exception_Name( The_Error ) &
                       " in " & Module & ".Factory_Task main loop" );
               raise;
         end;
      end loop;

      Leave( This.Debug, "" );

   exception
      when The_Error : Others =>
         for I in 1..Max_Agents loop
            if ACL(I).Process /= null then
               abort ACL(I).Process.all;
            end if;
         end loop;
         Error( This.Debug,
             "*** Exception " & Exception_Name( The_Error ) &
             " in " & Module & ".Factory_Task" );
         Leave( This.Debug, "some exception");
         raise;
   end Factory_Task;

   ---======================================================================---
   ---== VISIBLE METHODS ===================================================---
   ---======================================================================---

   --|
   --| Initialize the server component by allocating object
   --| instance space and startup the factory task.
   --| This procedure returns after the factory task has
   --| completed it's initialization process.
   --|
   procedure Initialize( O     : in Reference;
                         Port  : in Integer := 0;
                         debug : in Debugging_Support.Handle := null )
   is
      This              : Object_Data_Access := null;
      Inet_Socket_Name  : Internet_Socket_Address_Pointer
                                 := new Internet_Socket_Address;

      function To_Socket_PTR is
           new Unchecked_Conversion (
                        Source => Internet_Socket_Address_Pointer,
                        Target => Socket_Address_Pointer );

   begin
      Enter( debug, Module & ".Initialize" );

      if O.Data = null then
         O.Data := new Object_Data;
      end if;

      This       := O.Data;
      This.Self  := O;

      This.debug := debug;
      This       := O.Data;

      if Port > 0  then
         This.Port_Number := Internet_Port( Port );
      end if;

      Trace( This.Debug, "Create Socket");
      This.Listening_Socket := Create (Internet_Protocol, Stream_Socket);

      Trace( This.Debug, "Setting up IP & Port Address ");
      Set_Internet_Address (Inet_Socket_Name.all, Unspecified_Internet_Address);
      Set_Internet_Port ( Inet_Socket_Name.all, This.Port_Number);

      Trace( This.Debug, "Binding port & socket ");
      Bind (This.Listening_Socket, To_Socket_PTR( Inet_Socket_Name )  );

      Trace( This.Debug, "Setup Listen");
      Listen (This.Listening_Socket, 1);

      Trace( This.Debug, "IPC Setup complete");

      This.Factory := new Factory_Task( This );
      Leave( debug, "");

   exception
      when The_Error : Others =>
         Error( This.Debug,
             "*** Exception " & Exception_Name( The_Error ) &
             " in " & Module & ".Initialze" );
         Leave( This.Debug, " due to exception");
         raise;
   end Initialize;

   --|
   --| Finalize the factory by aborting the factory task
   --| and returning the object instance space.
   --|
   procedure Finalize( O : in Reference ) is
      This  : Object_Data_Access       := O.Data;
      debug : Debugging_Support.Handle := This.Debug;

      procedure Free is
            new Unchecked_Deallocation( Object_Data, Object_Data_Access);
   begin
      if This = null then
         return;
      end if;
      Enter( debug, Module & ".Finalize");

      abort This.Factory.all;
      Free( This );
      O.Data := null;

      Leave( debug, "");
   end Finalize;

end ASCL.SV.Server;
